<?php

namespace Drupal\nostr_content_nip23\Services;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Site\Settings;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Psr\Log\LoggerInterface;
use swentel\nostr\Event\Event;
use swentel\nostr\Message\EventMessage;
use swentel\nostr\Relay\Relay;
use swentel\nostr\Sign\Sign;

/**
 * Nostr publish event class.
 */
class NostrPublish implements NostrPublishInterface {

  use StringTranslationTrait;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected AccountInterface $currentUser;

  /**
   * The Messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected MessengerInterface $messenger;

  /**
   * Nostr storage service.
   *
   * @var \Drupal\nostr_content_nip23\Services\NostrStorageInterface
   */
  protected NostrStorageInterface $nostrStorage;

  /**
   * A logger instance.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected LoggerInterface $logger;

  /**
   * Show Drupal message or not.
   *
   * @var bool
   */
  protected bool $showDrupalMessage = FALSE;

  /**
   * NostrPublish constructor.
   *
   * @param \Drupal\Core\Session\AccountInterface $current_user
   *   The current user.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger.
   * @param \Drupal\nostr_content_nip23\Services\NostrStorageInterface $nostr_storage
   *   The nostr storage.
   */
  public function __construct(AccountInterface $current_user, MessengerInterface $messenger, LoggerInterface $logger, NostrStorageInterface $nostr_storage) {
    $this->currentUser = $current_user;
    $this->nostrStorage = $nostr_storage;
    $this->messenger = $messenger;
    $this->logger = $logger;
  }

  /**
   * {@inheritdoc}
   */
  public function sendMessage(EntityInterface $entity, string $public_key, string $private_key, string $relay, bool $show_drupal_message) {
    $this->showDrupalMessage = $show_drupal_message;

    // Get content from a field.
    $content = NULL;
    $content_fields = Settings::get('nostr_longform_content_field', ['body']);
    foreach ($content_fields as $content_field) {
      if ($entity->hasField($content_field)) {
        $content = trim($entity->get($content_field)->value);
        break;
      }
    }

    // @todo validate the content string

    if (!isset($content)) {
      $this->showDrupalMessage($this->t('No content found to send to the Nostr network.'), 'warning');
      $this->logger->warning('No content found in the fields to send to the Nostr network for @entity_type_id @entity_id', [
        '@entity_type_id' => $entity->getEntityTypeId(),
        '@entity_id' => $entity->id(),
      ]);
      return;
    }

    /**
     * Example Event
     * See https://github.com/nostr-protocol/nips/blob/master/23.md#example-event
     * {
     *  "kind": 30023,
     *  "created_at": 1675642635,
     *  "content": "Lorem [ipsum][nostr:nevent1qqst8cujky046negxgwwm5ynqwn53t8aqjr6afd8g59nfqwxpdhylpcpzamhxue69uhhyetvv9ujuetcv9khqmr99e3k7mg8arnc9] dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.\n\nRead more at nostr:naddr1qqzkjurnw4ksz9thwden5te0wfjkccte9ehx7um5wghx7un8qgs2d90kkcq3nk2jry62dyf50k0h36rhpdtd594my40w9pkal876jxgrqsqqqa28pccpzu.",
     *  "tags": [
     *    ["d", "lorem-ipsum"],
     *    ["title", "Lorem Ipsum"],
     *    ["published_at", "1296962229"],
     *    ["t", "placeholder"],
     *    ["e", "b3e392b11f5d4f28321cedd09303a748acfd0487aea5a7450b3481c60b6e4f87", "wss://relay.example.com"],
     *    ["a", "30023:a695f6b60119d9521934a691347d9f78e8770b56da16bb255ee286ddf9fda919:ipsum", "wss://relay.nostr.org"]
     *  ],
     *  "pubkey": "...",
     *    "id": "..."
     * }
     */

    $event = new Event();
    $event->setKind(30023);
    // If this content is not published, change the kind to 30024 (long form draft)
    $event->setPublicKey($public_key);
    $event->setTags([
      [ 'title', $entity->get('title')->value ],
      // @todo image tag
      // @todo summary tag
      [ 'published_at', sprintf('%s', $entity->get('created')->value) ],
      // @todo alt
      [ 'r', $relay ],
      [ 'L', '#t' ],
      [ 'l', '#Drupal', '#t' ],
      [ 'p', $public_key, $relay ]
    ]);
    if ($event_id = $this->nostrStorage->getEventId($entity->id(), $entity->getEntityTypeId())) {
      $event->setId($event_id);
    }
    $event->setContent($content);

    // Sign the event.
    $signer = new Sign();
    $alias_cleaner = \Drupal::service('pathauto.alias_cleaner');
    $event->addTag([ 'd', sprintf('%s', $alias_cleaner->cleanString($entity->get('title')->value)) ]);
    $signer->signEvent($event, $private_key);
    $eventMessage = new EventMessage($event);

    /**
     * Publish to these predefined relays.
     * Relays which support NIP-23.
     * - source: https://blogstack.io/naddr1qqyrjcnpxsmrgcesqy28wumn8ghj7un9d3shjtnyv9kh2uewd9hsygrxl93dnvvmdlsftgadpcc9xsjgnmx958p6r0qpxjlanz9w2snvwqpsgqqqw4rsq27597
     */
    $nip23_supported_relays = [
      'wss://relay.damus.io/',
      'wss://nos.lol',
      'wss://relay.plebstr.com',
      'wss://relay.nostr.band',
      'wss://nostr-01.yakihonne.com',
      'wss://nostr-02.yakihonne.com',
      'wss://nostr-02.dorafactory.org'
    ];
    $envelops = [];
    foreach ($nip23_supported_relays as $pre_def_relay) {
      $envelops[] = new Relay($pre_def_relay, $eventMessage);
    }

    // Publish to user defined relay
    $envelops[] = new Relay($relay, $eventMessage);

    try {
      /** @var Relay $envelop */
      foreach ($envelops as $envelop) {
        if ($this->currentUser->hasPermission('view nostr debugging')) {
          $this->showDrupalMessage(print_r('Sending event to relay ' . $envelop->getUrl(), 1), 'warning');
          $this->showDrupalMessage($eventMessage->generate(), 'warning');
        }
        /** @var \swentel\nostr\Relay\CommandResult $response */
        $response = $envelop->send();

        // View debugging information.
        if ($this->currentUser->hasPermission('view nostr debugging')) {
          $this->showDrupalMessage(print_r($response, 1), 'warning');
        }

        if ($response->isSuccess() === FALSE) {
          $failure = $response->message();
          $this->showDrupalMessage($this->t('Failed posting to Nostr network: @fail', ['@fail' => $failure]), 'error');
          $this->logger->error('Failed posting to Nostr network: @message', ['@message' => $failure]);
        } else {
          // Store the id.
          if (Settings::get('nostr_nip23_sent_nids', TRUE)) {
            $this->nostrStorage->storeEventId($entity->id(), $entity->getEntityTypeId(), $event->getId(), $envelop->getUrl());
          }
        }
      }
    }
    catch (\Exception $e) {
      $this->showDrupalMessage($this->t('Exception error: @message', ['@message' => $e->getMessage()]), 'error');
      $this->logger->error('Exception error: @message', ['@message' => $e->getMessage()]);
    }

  }

  /**
   * Helper function to show a Drupal message.
   *
   * @param string $message
   *   The message.
   * @param string $type
   *   Type of message.
   */
  protected function showDrupalMessage(string $message, string $type = 'default') {
    if ($this->showDrupalMessage) {
      switch ($type) {
        case 'error':
          $this->messenger->addError($message);
          break;

        case 'warning':
          $this->messenger->addWarning($message);
          break;

        default:
          $this->messenger->addMessage($message);
          break;
      }
    }
  }

}
